######################################################################
#                                                                    #
#                                                                    #
# Dfm2gui: Converts Delphi/C++Builder Forms to Win32::GUI templates. #
#                                                                    #
# Usage  : dfm2gui input.dfm [output.pl]                             #
# Note   : This script does not work with the binary version of DFM  #
#          files. It attempts to use the external dfm2txt utility to #
#          convert binary forms to text prior to parsing. dfm2txt    #
#          should be supplied as part of this distribution.          #
# Dfm2gui version: 0.024                                             #
# Dfm2gui date   : 13/5/2000                                         #
# Bugs/comments  : writeexcel@eircom.net                             #
#                                                                    #
# Copyright (c) 2000, John McNamara. Under the same terms as Perl.   #
#                                                                    #
#                                                                    #
######################################################################

#!perl -w

use strict;
#use diagnostics;
use MIME::Base64;
#use Data::Dumper;


######################################################################
#
# Global variables
#
my $version      = 0.024;
my $verdate      = '13/5/2000';
my $email        = 'writeexcel@eircom.net';
my $obj_level    = -1;
my $menu_level   = -1;
my $save_bitmaps = 1;
my $mainform;
my $mainmenu;
my @constructors;
my @obj;
my @menus;
my @popups;
my @myvars;
my @events;
my @bitmaps;
my %TControl;

my $propwidth    = 14; # Default width for the key of GUI controls
my $commentwidth = 70; # Width of comment lines



######################################################################
#
# File opening routines
#


######################################################################
# Check the number of Command line arguments
#
die print_usage() if ((@ARGV < 1) || (@ARGV > 2));

my $dfmfile = $ARGV[0];
my $guifile = $ARGV[1];
my $tmpfile = '';


######################################################################
# First arg should be Delphi/C++Builder *.dfm file.
#
open (DELPHI, $dfmfile) or die "\nDFM file $dfmfile not found.\n";


######################################################################
# Check if DFM file is in binary format. Use dfm2txt to convert it to
# a temorary text file if required.
#
if (-B $dfmfile) {
    $tmpfile = "~tmpdfm.txt";
    system(dfm2txt, $dfmfile, $tmpfile);
    if ($? == 0) {
        open (DELPHI, $tmpfile)
                     or die "\nCannot open temporary file $dfmfile\n";
    }
    else {
        die "\nDfm2txt utility not found to convert binary DFM file. "
            . "Use dfm2txt utility \nto convert $dfmfile to text.\n";
    }
}


######################################################################
# If second arg is not specified use the filename of the first arg
# and substitute the extension.
#
if (@ARGV == 2) {
    open (GUISCRIPT,">$guifile") or die "\nCan't open $guifile.\n";
}
else {
    $guifile = $dfmfile;
    if ($guifile =~ s/.dfm$/.pl/i) {
       open (GUISCRIPT,">$guifile") or die "\nCan't open $guifile.\n";
    }
    else { die print_usage() }
}


######################################################################
#
# Main routines
#
#
print "\nConverting form $ARGV[0] to Win32::GUI template $guifile.\n";
initialize_TControl();
read_delphi_objects();
parse_delphi_objects();

print_script_header();
print_myvars();
print_bitmap_initializer();
print_menu();
print_popups();
print_constructors();
print_main_initializer();
print_events();
print_inline_bitmaps();
print "Conversion finished.\n\n";
if ($tmpfile) {close DELPHI; system(del, $tmpfile)}


# Some debugging code
#
#print Dumper(\%TControl);
#open (TXT, ">z.dfm");
#print TXT @obj;




######################################################################
#
# Print usage message.
#
sub print_usage {

"
Dfm2gui: Converts Delphi/C++Builder Forms to Win32::GUI templates.
         Copyright, John McNamara, $email.
         Version $version, $verdate.


Usage  : dfm2gui input.dfm [output.pl]
Example: dfm2gui Myform.dfm Mygui.pl

Options: [output] If no output file is specified then the filename of
                  the input file is used and the extension is changed.



Note   : This script does not work with the binary version of DFM
         files. It attempts to use the external dfm2txt utility to
         convert binary forms to text prior to parsing. dfm2txt
         should be supplied as part of this distribution.

"
}


######################################################################
#
# Prints a box of #'s to separate sections in the Win32::GUI script.
#
sub commentbox {
    my $outstring;
    my $line;
    my @comments = @_;

    unshift @comments, ""; # Add blank line to beginning
    push    @comments, ""; # Add blank line to end

    $outstring = "#" x $commentwidth ."\n";
    foreach $line (@comments) {
        $outstring .= sprintf ( "# %-*s #\n", $commentwidth-4, $line);
    }
    return $outstring .= "#" x $commentwidth ."\n";
}


######################################################################
#
# Read the objects in the dfm file. The read is recursive to keep
# track of parent objects. Objects on each level are stored in @obj.
# This sub spawns sub subroutines to deal with:
# * Main menus
# * Popup menus
# * Inline bitmaps, glyphs and icons
#
sub read_delphi_objects {
    my $parent;
    my $child;
    my $line;
    my $type;
    my $left = 0;
    my $top  = 0;

    my $parent_left;
    my $parent_top;

    if (@_ != 0) {
        $obj_level++;
        $parent_left = shift @_;
        $parent_top  = shift @_;
        $parent      = shift @_;
        $line        = shift @_;
        (undef, $child, $type ) = split(' ', $line, 3);
        chop($child);
        $obj[$obj_level]  .= "  " x $obj_level . "$parent $line";
     }
     else {
        $parent_left = 0;
        $parent_top  = 0;
        $child = 'None';
     }


    while (<DELPHI>) {
        my $tmp  = $_;

        if (/TMainMenu$/) { read_delphi_menu();    next}
        if (/TPopupMenu$/){ read_delphi_popup($_); next}
        if (/Bitmap/)     { read_delphi_bitmap($child) }
        if (/Glyph\.Data/){ read_delphi_glyph($child)  }
        if (/Icon\.Data/) { read_delphi_icon($child)   }


        if ((/^\s*Left/) && ($obj_level >0)) {
            (undef, undef, $left ) = split(' ', $_, 3);
            chop($left);
            $left += $parent_left;
            $tmp = "  " x ($obj_level+1) . "Left = $left\n";
        }

        if ((/^\s*Top/) && ($obj_level >0)) {
            (undef, undef, $top ) = split(' ', $_, 3);
            chop($top);
            $top += $parent_top;
            $tmp = "  " x ($obj_level+1) . "Top = $top\n";
        }

        if (/^\s*end$/) {
            $obj[$obj_level] .= $tmp;
            $child = $parent;
            $obj_level--;
            return;
        }

        if (/^\s*object/) {
            read_delphi_objects($left, $top, $child, $_);
            next;
        }

        $obj[$obj_level] .= $tmp;
    }
}


######################################################################
#
#  Read main menus.
#
sub read_delphi_menu {
    my $menu_item;
    my $caption;
    my $property;

    while (<DELPHI>) {
        if (/TMenuItem$/) {
            $menu_level++;
            (undef, $menu_item, undef ) = split(' ', $_, 3);
            chop($menu_item);
            next;
        }

        if (/^\s*Caption/) {
            (undef, undef, $caption) = split(' ', $_, 3);
            chop($caption);
            my $indent = '  ' x $menu_level . '>' x $menu_level;
            substr($caption, 1, 0, $indent);
            push @menus, {$caption => $menu_item};
            next;
        }

        if (/^\s*On/) {
            ($property, undef, undef) = split(' ', $_, 3);
            if (exists $TControl{'TComponent'}{$property}) {
                push @events, $menu_item .
                                   $TControl{'TComponent'}{$property};
            }
            next;
        }

        if (/^\s*end$/) {
            if ($menu_level == -1) {return}
            $menu_level--;
        }
    }
}


######################################################################
#
#  Read popup menus.
#
sub read_delphi_popup {
    my $menu_item;
    my $caption;
    my $line = shift @_;

    $menu_level++;
    (undef, $menu_item, undef ) = split(' ', $line, 3);
    chop($menu_item);
    push @popups, {'' => $menu_item};
    push @myvars, $menu_item;

    while (<DELPHI>) {
        if (/TMenuItem$/) {
            $menu_level++;
            (undef, $menu_item, undef ) = split(' ', $_, 3);
            chop($menu_item);
        }

        if (/^\s*Caption/) {
            (undef, undef, $caption) = split(' ', $_, 3);
            chop($caption);
            my $indent = '  ' x $menu_level . '>' x $menu_level;
            substr($caption, 1, 0, $indent);
            push @popups, {$caption => $menu_item};
        }

        if (/^\s*end$/) {
            $menu_level--;
            if ($menu_level == -1) {return}
        }
    }
}


######################################################################
#
# Read and convert Delphi inline bitmap to Win32::GUI inline bitmap.
#
sub read_delphi_bitmap {
    my $parent = shift @_;
    my $data = read_delphi_image();
    my $bitmap;
    my $mask;
    (undef, $bitmap, $mask) = split (/424D/,$data, 3);
    $bitmap = '424D' . $bitmap;
    $mask   = '424D' . $mask;

    inline_image($parent . "_bmp", $bitmap, 0);
    inline_image($parent . "_msk", $mask, 0);
}


######################################################################
#
# Read and convert Delphi inline icon to Win32::GUI inline bitmap.
#
sub read_delphi_icon {
    my $parent = shift @_;
    my $data = read_delphi_image();
    inline_image($parent, $data, 1);
}


######################################################################
#
# Read and convert Delphi inline glyph to Win32::GUI inline bitmap.
#
sub read_delphi_glyph {
    my $parent = shift @_;
    my $data = read_delphi_image();
    $data = substr ($data, 8);
    inline_image($parent, $data, 0);
}


######################################################################
#
# Read Delphi inline image and return a hex string.
#
sub read_delphi_image {
    my $data = '';

    while (<DELPHI>) {
        chop;
        s/^\s+//;       # Strip leading space
        $data .= $_;
        if (/}/) {      # Match closing brace of
            chop($data);# Chop }
            last;
        }
    }
    return $data;
}


######################################################################
#
# Convert Delphi inline image to GUI inline images.
#
sub inline_image {

    my $name    = $_[0]; # Parent name
    my $data    = $_[1]; # Delphi hex data
    my $type    = $_[2]; # 0 = bmp, 1 = ico
    my $hex_str;
    my $bin_str;
    my $mim_str;

    my $save_ok = $save_bitmaps; # Save image to disk?
    if ($save_ok) {
        my $filename;

        if ($type == 0) { # Save bmp or ico
            $filename = $name . ".bmp";
        }
        else {
            $filename = $name . ".ico"
        }
        # Try to open image file
        open (BITMAP,  ">$filename") or ($save_ok = 0);
        if ($save_ok) {
            binmode(BITMAP);
            print "Extracting $filename.\n";

        }
        else {
            print STDERR "Cannot write bitmap to $filename.\n";
        }
    }

    # Store BitmapInline constructor
    push @myvars, "BMP_$name";
    push @bitmaps, "\n    \$BMP_$name = new Win32::GUI::BitmapInline(\n";
    push @bitmaps, "        q{\n";

    # 1. Chop hex string into lengths of 57 bytes, see MIME::Base64
    # 2. Pack the string as binary data
    # 3. Encode the binary string as Mime
    while ($data) {
        $hex_str = substr($data, 0, 57*2, ''); # 2 chars per byte
        $bin_str = pack("H*", $hex_str);
        $mim_str = "        " . MIME::Base64::encode($bin_str);
        push @bitmaps, $mim_str;
        if ($save_ok) { print BITMAP $bin_str}
    }

    push @bitmaps, "        }\n";
    push @bitmaps, "    );\n";
    close BITMAP;
}


######################################################################
#
# Convert the objects in @obj into Win32::GUI controls
#
sub parse_delphi_objects {
    while (@obj) {
        my $classexists;
        my $line;
        my $parent;
        my $property;
        my $object;
        my $type;
        my $key;
        my $value;
        my @lines = split(/\n/, shift @obj);

        foreach $line (@lines) {
            # Match start of object description
            if ($line =~ /\s*object/) {
                ($parent, undef, $object, $type) = split(' ',$line,4);
                chop($object);
                if ($parent eq 'None') {
                    $type   = 'TForm';
                    $mainform = $object;
                }
                if (exists $TControl{$type}) {
                    push @myvars, $object;
                    my $constructor = $TControl{$type}{'Constructor'};
                    store_constructor($mainform, $constructor);
                    store_property('-name', "'$object'");
                    $classexists = 1;
                }
                else {
                    $classexists = 0;
                }
            }

            # Match object properties
            if (($classexists) and ($line =~ /=/)) {
                ($property, undef, $value) = split(' ', $line, 3);
                if ($property eq 'Menu') {
                    push @myvars, $value;
                    $mainmenu = $value;
                    $value = '$' . $value;
                }
                if ($property =~ /^On/) {
                    if (exists $TControl{'TComponent'}{$property}) {
                        push @events, $object .
                                   $TControl{'TComponent'}{$property};
                    }
                    next;
                }
                if ($value eq 'False') {$value = 0}
                if ($value eq 'True' ) {$value = 1}
                if (exists $TControl{'TComponent'}{$property}) {
                    store_property($TControl{'TComponent'}{$property},
                                                           $value);
                }
                if (exists $TControl{$type}{$property}) {
                    store_property($TControl{$type}{$property},
                                                    $value);
                }
            }

            # Match end of object description
            if ($classexists and ($line =~ /^\s*end$/)) {
                if (exists $TControl{$type}{'_required'}) {
                    while (($key,$value) =
                              each %{$TControl{$type}{'_required'}}) {
                        store_property($key, $value);
                    }
                }
                push @constructors, ");\n\n";
             }
         }
    }
}


######################################################################
#
# Store Win32::GUI package constructors in @constructors
#
sub store_constructor {
    my $parent      = shift @_;
    my $constructor = shift @_;
    push @constructors, "\$$parent$constructor(\n";
}


######################################################################
#
# Store Win32::GUI package properties in @constructors.
#
sub store_property {
    my $property = shift @_;
    my $value    = shift @_;
    $value .= ',';
    push @constructors, sprintf("    %-*s => %-s\n",
                                       $propwidth, $property, $value);
}


######################################################################
#
# Print constructors to Win32::GUI script.
#
sub print_constructors {
    print GUISCRIPT @constructors;
}


######################################################################
#
# Print "my" vars at beginning of Win32::GUI script to comply
# with "use strict".
#
sub print_myvars {
    foreach my $var (@myvars) {print GUISCRIPT "my \$$var;\n"}
    print GUISCRIPT "\n";
}


######################################################################
#
# Print main menus to Win32::GUI script.
#
sub print_menu {
    if (@menus == 0) {return}
 
    print GUISCRIPT "\$$mainmenu = new Win32::GUI::Menu(\n";

    my $width = $propwidth;

    foreach my $menu (@menus) {
        foreach my $caption ( keys %$menu ) {
            my $len = length($caption);
            if ($len > $width) {$width = $len};
        }
    }

    foreach my $menu (@menus) {
        foreach my $caption ( keys %$menu ) {
            my $name = $menu->{$caption};
            $name = "'$name',";
            printf GUISCRIPT ("    %-*s => %-s\n",
                                             $width, $caption, $name);
        }
    }

    print GUISCRIPT ");\n\n";
}


######################################################################
#
# Print pop-up menus to Win32::GUI script.
#
sub print_popups {
    if (@popups == 0) {return}

    my $width = $propwidth;
    my $firstpopup = 1;

    foreach my $menu (@popups) {
        foreach my $caption ( keys %$menu ) {
            my $len = length($caption);
            if ($len > $width) {$width = $len};
        }
    }

    foreach my $menu (@popups) {
        foreach my $caption ( keys %$menu ) {
            my $name = $menu->{$caption};
            if ($caption eq '') {
                if ($firstpopup != 1) {print GUISCRIPT "\n);\n\n"}
                $firstpopup = 0;
                print GUISCRIPT "\$$name = new Win32::GUI::Menu(\n";
                $caption = "''";
            }

            $name = "'$name',";
            printf GUISCRIPT ("    %-*s => %-s\n",
                                             $width, $caption, $name);
        }
    }

    print GUISCRIPT ");\n\n";
}


######################################################################
#
# Print header and shebang line to Win32::GUI script.
#
sub print_script_header {
    my $outstring = commentbox(
        "",
        "Win32::GUI template produced by Dfm2gui.pl",
        "" ,
        "Dfm2gui version: $version",
        "Dfm2gui date   : $verdate",
        "Bugs/comments  : $email",
        ""
       );

    print GUISCRIPT $outstring;

    $outstring = "\n\n\n" . commentbox("Main") . "\n";
    $outstring .=  "#!perl -w\n\n\n";
    $outstring .=  "use strict;\n";
    $outstring .=  "use Win32::GUI;\n";
    if (@bitmaps){ $outstring .= "use Win32::GUI::BitmapInline;\n" }
    $outstring .=  "\n";


    print GUISCRIPT $outstring;
}


######################################################################
#
# Print window initializer to Win32::GUI script.
#
sub print_main_initializer {
    my $outstring =
        "\$$mainform->Show;\n" .
        "Win32::GUI::Dialog(0);\n\n\n";

    print GUISCRIPT $outstring;
}


######################################################################
#
# Print events sections of Win32::GUI script.
# Adds a dummy message dialog for testing events.
#
sub print_events {
    my $outstring =
            commentbox("Events") . "\n" .
            "sub ${mainform}_Terminate {\n" .
            "    if (\$$mainform) {\$$mainform->Close()}\n" .
            "    exit;\n" .
            "}\n";

    print GUISCRIPT $outstring;

    foreach my $event (@events) {
        $outstring =
            "\nsub $event {\n    ".
            "Win32::GUI::MessageBox(\$$mainform, \"$event\", ".
            "\n                           ".
            '"Event handler", MB_ICONINFORMATION);'.
            "\n}\n";
        print GUISCRIPT $outstring;
    }

}


######################################################################
#
# print the Bitmap section of the Win32::GUI script.
#
sub print_inline_bitmaps {
    if (@bitmaps == 0) {return}

    print GUISCRIPT "\n\n", commentbox("Bitmaps") ;
    print GUISCRIPT "\nsub initialize_bitmaps {\n";
    print GUISCRIPT @bitmaps;
    print GUISCRIPT "}\n";
}


######################################################################
#
# print initialize_bitmaps() to the Win32::GUI script.
#
sub print_bitmap_initializer {
    if (@bitmaps == 0) {return}

    print GUISCRIPT "\ninitialize_bitmaps();\n\n";
}


######################################################################
#
# Initialize the main hash that stores the relationships between
# Delphi controls and Win32::GUI controls.
#
sub initialize_TControl {

    %TControl = (

        TComponent => {
            Left           => '-left',
            Top            => '-top',
            Width          => '-width',
            Height         => '-height',
            Caption        => '-text',
            OnClick        => '_Click',
            OnEnter        => '_GotFocus',
            OnExit         => '_LostFocus',
            OnMouseDown    => '_MouseDown',
            OnMouseUp      => '_MouseUp',
        },

        TAnimate => {
            Constructor    => '->AddAnimation',
        },

        TButton => {
            Constructor    => '->AddButton',
            Cancel         => '-cancel',
        },

        TCheckBox => {
            Constructor    => '->AddCheckbox',
        },

        TComboBox => {
            Constructor    => '->AddCombobox',
        },

        TEdit => {
            Constructor    => '->AddTextfield',
            Text           => '-text',
            ReadOnly       => '-readonly',
            PasswordChar   => '-passwordchar',
        },

        TForm => {
            Constructor    => ' = new Win32::GUI::Window',
            Menu           => '-menu',
            _required      => {
                              '-visible'    => 1,
            },
        },

        TGroupBox => {
            Constructor    => '->AddButton',
            _required      => {
                              '-style' => "WS_VISIBLE | WS_CHILD | 7",
            },
        },

        TLabel => {
            Constructor    => '->AddLabel',
            WordWrap       => '-wrap',
        },

        TListView => {
            Constructor    => '->AddListView',
        },

        TMemo => {
            Constructor    => '->AddTextfield',
            Text           => '-text',
            ReadOnly       => '-readonly',
            PasswordChar   => '-passwordchar',
            _required      => {
                               '-multiline '    => 1,
            },
        },

        TProgressBar => {
            Constructor    => '->AddProgressBar',
            Smooth         => '-smooth',
        },

        TRadioGroup => {
            Constructor    => '->AddButton',
            _required      => {
                              '-style' => "WS_VISIBLE | WS_CHILD | 7",
            },
        },

        TRadioButton => {
            Constructor    => '->AddRadioButton',
        },

        TRichEdit => {
            Constructor    => '->AddRichEdit',
        },

        TStatusBar => {
            Constructor    => '->AddStatusBar',
        },

        TToolBar => {
            Constructor    => '->AddToolbar',
        },

        TTrackBar => {
            Constructor    => '->AddSlider',
        },

        TTreeView => {
            Constructor    => '->AddTreeView',
        },

        TUpDown => {
            Constructor    => '->AddUpDown',
        },

    );
}
